#@+leo
#@+node:0::@file open_shell.py
#@+body
#@@language python

"""Opens up cmd and explorer window to same directory 
as @file nodes or children ..."""


#@<< about the open shell plugin >>
#@+node:1::<< about the open shell plugin >>
#@+body
#@+at
#  Written by Ed Taekema.  Modified by E.K.Ream
# 
# Please submit bugs / feature requests to etaekema@earthlink.net"""
# 
# This is a simple plugin for leo 3.12 that allows the user to open either an 
# xterm on linux or a cmd windows/explorer window on win32 in the directory of 
# the current @file.  This allows quick navigation to facilitate testing and 
# navigating large systems with complex direcgtories.
# 
# Current limitations ...
# 
# 1. Not tested on Mac OS X ...
# 2. On win32, the cmd window will not open in the right directory if the 
# @file location is on a different drive than the .leo file that is being edited.
# 3. On linux, xterm must be in your path.

#@-at
#@-body
#@-node:1::<< about the open shell plugin >>


import leo,os,os.path,sys
from leoPlugins import *
from leoGlobals import *

pathToExplorer = 'c:/windows/explorer.exe'
pathToCmd = 'c:/windows/system32/cmd.exe'


#@+others
#@+node:2::load_menu
#@+body
def load_menu(tag,keywords):
	
	if sys.platform=="win32":
		table = (
			("&Open Console Window",None,launchCmd),
			("Open &Explorer",None,launchExplorer)) 
	else:
		table = ( ("Open &xterm",None,launchxTerm), ) 
	
	top().frame.createNewMenu("E&xtensions","top")
	top().frame.createMenuItemsFromTable("Extensions",table)

#@-body
#@-node:2::load_menu
#@+node:3::_getpath
#@+body
def _getpath(c,v):
	dict = scanDirectives(c,v)
	d = dict.get("path")

	if d == None:

		if v.isAtFileNode():
			filename = v.atFileNodeName()
		if v.isAtNoSentinelsFileNode():
			filename = v.atNoSentinelsFileNodeName()
		if v.isAtRawFileNode():
			filename = v.atRawFileNodeName()
		if v.isAtSilentFileNode():
			filename = v.atSilentFileNodeName()

		d = os.path.dirname(filename)

	d = os.path.normpath(d)
	return d

#@-body
#@-node:3::_getpath
#@+node:4::_getcurrentnodepath
#@+body
def _getCurrentNodePath():
	c = leo.top()
	v = c.currentVnode()
	f = v.atFileNodeName()
	d = _getpath(c,v)
	return d

#@-body
#@-node:4::_getcurrentnodepath
#@+node:5::launchCmd
#@+body
def launchCmd(not_used):
	
	global pathToCmd

	d = _getCurrentNodePath()
	myCmd = 'cd ' + d
	os.spawnl(os.P_NOWAIT, pathToCmd, '/k ', myCmd)
#@-body
#@-node:5::launchCmd
#@+node:6::launchExplorer
#@+body
def launchExplorer(not_used):
	
	global pathToExplorer

	d = _getCurrentNodePath()
	os.spawnl(os.P_NOWAIT,pathToExplorer, ' ', d)


#@-body
#@-node:6::launchExplorer
#@+node:7::launchxTerm
#@+body
def launchxTerm(not_used):

	d = _getCurrentNodePath()
	curdir = os.getcwd()
	os.chdir(d)
	os.spawnlp(os.P_NOWAIT, 'xterm', '-title Leo')
	os.chdir(curdir)
#@-body
#@-node:7::launchxTerm
#@-others


if 0: # Register the plugin

	registerHandler("after-create-leo-frame", load_menu)
	__version__ = "1.3"
	plugin_signon(__name__)
#@-body
#@-node:0::@file open_shell.py
#@-leo
