#@+leo
#@+node:0::@file scripts_menu.py
#@+body
"""Create a Scripts menu for LeoPy.leo"""

from leoPlugins import *
from leoGlobals import *
import glob,os


#@+others
#@+node:1::createScriptsMenu
#@+body
def createScriptsMenu (tag,keywords):

	c = keywords.get("c")
	path = os.path.join(app().loadDir,"..","scripts")
	
	if os.path.exists(path):

		# Create lists of scripts and subdirectories.
		entries = glob.glob(os.path.join(path,"*"))
		top_scripts = glob.glob(os.path.join(path,"*.py"))
		dirs = [f for f in entries if os.path.isdir(f)]
		
		#@<< Return if no scripts exist anywhere >>
		#@+node:1::<< Return if no scripts exist anywhere >>
		#@+body
		if not top_scripts:
			found = false
			for dir in dirs:
				scripts = glob.glob(os.path.join(dir,"*.py"))
				if scripts:
					found = true ; break
			if not found:
				return
		
		#@-body
		#@-node:1::<< Return if no scripts exist anywhere >>

	
		scriptsMenu = c.frame.createNewMenu("&Scripts")
		table = []
		
		#@<< Create top-level entries for every script in top_scripts >>
		#@+node:2::<< Create top-level entries for every script in top_scripts >>
		#@+body
		table = []
		top_scripts.sort()
		for script in top_scripts:
			name = shortFileName(script)
			def doScript(event=None,name=name):
				executeScript(name)
			table.append((name,None,doScript),)
			
		c.frame.createMenuEntries(scriptsMenu, table)
		
		#@-body
		#@-node:2::<< Create top-level entries for every script in top_scripts >>

		for dir in dirs:
			files = glob.glob(os.path.join(dir,"*.py"))
			if files:
				
				#@<< Create a submenu for dir containing each file in files >>
				#@+node:3::<< Create a submenu for dir containing each file in files >>
				#@+body
				# Create the submenu.
				name = os.path.join("scripts",shortFileName(dir))
				menu = c.frame.createNewMenu(name,"&Scripts")
				
				# Populate the submenu.
				table = []
				for file in files:
					name = shortFileName(file)
					def doScript(event=None,name=name):
						from leoGlobals import executeScript
						executeScript(name)
					table.append((name,None,doScript),)
				
				c.frame.createMenuEntries(menu, table)
				#@-body
				#@-node:3::<< Create a submenu for dir containing each file in files >>
#@-body
#@-node:1::createScriptsMenu
#@-others


if 1: # Register the handlers...
	registerHandler("create-optional-menus",createScriptsMenu)
	
	__version__ = "1.4"
	plugin_signon(__name__)
#@-body
#@-node:0::@file scripts_menu.py
#@-leo
